 #!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Dec  6 14:45:16 2018

@author: zoerehder

Simple MeEP run file
"""

import sys
FLake_path    = "/absolute/path/to/FLake"

K_to_C = 273.15 # Kelvin = 0 Celsius

if FLake_path not in sys.path:
    sys.path.insert(0,FLake_path)

import flake_module as fl
import xarray as xr
import numpy as np
from FLake_parameter import *
from Methane.function_FLake_ice_plant_cy import methane_main
from Hydrology.hydrologie_main import hydrology
from SoilHeat.Cryo_forFlake_commonsurface_cy import CryoGridImplizit
from datetime import datetime as dt
from datetime import timedelta
np.seterr(divide='ignore', invalid='ignore')
from pandas import Timestamp
import pickle

#Samoylov setup
from SoilHeat.cryo_ini_Samoylov import initialize_soilheat
from Hydrology.topo_parameter_Samoylov import initialize_topo_hydro, set_overgrownfraction

#%% -- FORCING DATA ---

Forcing_Description = 'Source and time step length'
Run_Description     = 'Location, duration, ...'
Outfilename         = '/path/to/output/folder/'+dt.now().strftime('%y%m%d')+'_runname'

forcing = xr.open_dataset(historical_name+'_forcing.nc')

lai_max = forcing.lai.max()
lai_min = forcing.lai.min()

forc_snowfall   = forcing.tp.where(forcing.T_a.rolling(time=3, center=True).max()<273.15,0).values
forc_I_atm      = forcing.I_atm.values
forc_Q_atm_lw   = forcing.Q_atm_lw.values
forc_U_a        = forcing.U_a.values
forc_T_a        = forcing.T_a.values
forc_q_a        = forcing.q_a.values
forc_P_a        = forcing.P_a.values
forc_tp         = forcing.tp.values
forc_npp        = forcing.NPP.values
forc_lai        = forcing.lai.values
forc_time       = forcing.time.values


pondtype = np.array(['iw', 'pc','mp'])#
topography, hyd_output = initialize_topo_hydro(forcing.time.values,pondtype=pondtype)

topo_A          = topography.A.values
topo_alpha      = topography.alpha.values
topo_super_wb   = topography.super_wb.values
topo_super_rim  = topography.super_rim.values
topo_depth_wb   = topography.depth_wb.values
topo_waterbody  = topography.waterbody.values

hyd_A_wb    = hyd_output.A_wb.values
hyd_A_veg   = hyd_output.A_veg.values
hyd_A_bare  = hyd_output.A_bare.values
hyd_h_veg   = hyd_output.h_veg.values
hyd_h_bare  = hyd_output.h_bare.values
hyd_open_w  = hyd_output.open_w.values
hyd_WT_wb   = hyd_output.WT_wb.values

#%% -- SOIL HEAT --
Z = np.append(np.arange(-0.5, 0, 0.02), 
              np.append(np.arange(0,5, 0.05), 
                        np.append(np.arange(5,10, 0.2),
                                  np.arange(10,20.5,0.5))))
depth_w_ini = topography.super_wb.values[:]- hyd_output.WT_wb.values[:,0]
cr_GRID, cr_PARA, cr_STRAT, cr_STATVAR, cr_TEMP, cr_output \
          = initialize_soilheat(Z         = Z,
                                waterbody = topography.waterbody.values, 
                                timesteps = forcing.time.values,
                                pondtype = pondtype)
cr_T_soil = cr_output.T_soil.values
cr_W_soil = cr_output.W_soil.values
cr_I_soil = cr_output.I_soil.values
cr_TD     = cr_output.TD.values
cr_Q_bot  = cr_output.Q_bot.values
cr_WaterDepth = cr_PARA['WaterDepth'][0,:]
cr_snow_depth = cr_output.Snow_depth.values

# -- FLAKE --
datashape = hyd_A_wb.shape

fl_T_snow   = np.zeros(datashape)
fl_T_ice    = np.zeros(datashape)
fl_T_mnw    = np.zeros(datashape)
fl_T_wML    = np.zeros(datashape)
fl_T_bot    = np.zeros(datashape)
fl_C_T      = np.zeros(datashape)
fl_h_snow   = np.zeros(datashape)
fl_h_ice    = np.zeros(datashape)
fl_h_ML     = np.zeros(datashape)
fl_T_sfc_p  = np.zeros(datashape)
fl_w_dd     = np.zeros(datashape)
fl_u_star   = np.zeros(datashape)
fl_E        = np.zeros(datashape)
fl_I_bot    = np.zeros(datashape)

meth_cCH4_ml_v      = np.zeros(datashape)
meth_F_diff_surf_v  = np.zeros(datashape)
meth_F_bubble_v     = np.zeros(datashape)
meth_F_plant        = np.zeros(datashape)
meth_cCH4_ml_b      = np.zeros(datashape)
meth_F_diff_surf_b  = np.zeros(datashape)
meth_F_bubble_b     = np.zeros(datashape)
meth_cO2_ml         = np.zeros(datashape)
meth_stored_CH4     = np.zeros(datashape)

#-- SWITCH FOR METHANE MODULE
meth = True
 #%% Initialization for flake-physics
fl_C_T[:,0]     = C_T_in   
fl_h_ice[:,0]   = 1
fl_h_snow[:,0]  = h_snow_in      
fl_h_ML[:,0]    = 1   
fl_T_bot[:,0]   = T_bot_in
fl_T_ice[:,0]   = T_ice_in
fl_T_mnw[:,0]   = T_mnw_in
fl_T_sfc_p[:,0] = T_mnw_in
fl_T_snow[:,0]  = T_ice_in
fl_T_wML[:,0]   = T_wML_in
fl_w_dd[0]      = np.nan
fl_u_star[0]    = np.nan
stored_CH4      = 0
ice_CH4         = 0


#%% 
# seconds per timestep
del_time     = (forc_time[1]- forc_time[0]).astype('timedelta64[s]').astype('int') 
ice_timesteps = int(del_time/3600 * 24* 30) # timesteps of the last 200 days
tsteps_1day  = int(24*60*60/del_time)
air_idx      = np.max(np.where(cr_STATVAR['Zp']<0))
sediment_idx = np.sum(cr_STRAT['SolidMatrix'][:,:]==0, axis = 0)
por = np.zeros(len(sediment_idx))
for i,s in enumerate(sediment_idx):
    por[i] = 1 - cr_STRAT['SolidMatrix'][s,i]
#variables for resetting overgrown fraction
reset_now = np.zeros(len(hyd_output.waterbody))
lim_reset = int(del_time/24*5)
h_w_veg   = 0.5

ice_on = np.zeros(len(topo_waterbody))

# 1st timestep after spin-up for resetting overgrown fraction
reset_veg = int(np.argwhere(forc_time==np.Datetime64('2004-01-01')))


#%% ---------------------- LOOP -------------------------

for t in range(len(forc_time)-1):

    this_t = t
    # ----------------------------------------------
    
    if this_t % 480 == 0:
        print(forc_time[t])
#   ------ Soil heat ------    
    
    cr_T_soil[:,:,t+1]  ,\
    cr_W_soil[:,:,t+1]  ,\
    cr_I_soil[:,:,t+1]  ,\
    cr_Q_bot   [:,t+1]  ,\
    cr_TD      [:,t+1]  ,\
    cr_STRAT                   ,\
    cr_STATVAR                 ,\
    cr_TEMP                    ,\
    cr_snow_depth[t+1],\
        = CryoGridImplizit(                 
            forc_T_a[t],
            fl_I_bot[:, t],
            cr_GRID   ,
            cr_STRAT  ,
            cr_WaterDepth ,
            cr_STATVAR,
            cr_TEMP,
            Timestamp(forc_time[t]),
            forc_snowfall[this_t])    
        
 
    for p in range(len(topo_waterbody)):
        
        if (cr_I_soil[air_idx,0,this_t]>0) and (fl_h_ice[p,this_t]>0):
            T_flakeforcing = cr_T_soil[air_idx,0,this_t]
            u_flakeforcing = forc_U_a[this_t]  
            I_flakeforcing = forc_I_atm[this_t]
        else:
            T_flakeforcing = forc_T_a[this_t]
            u_flakeforcing = forc_U_a[this_t]  
            I_flakeforcing = forc_I_atm[this_t]
            
        depth_w = max(0,topo_super_wb[p]- hyd_WT_wb[p,t])
        if depth_w > 0.01:
#       ------ FLAKE PHYSICS ------   
            (fl_T_snow[p,t+1],  \
            fl_T_ice[p,t+1],    \
            fl_T_mnw[p,t+1],    \
            fl_T_wML[p,t+1],    \
            fl_T_bot[p,t+1],    \
            fl_C_T[p,t+1],      \
            fl_h_snow[p,t+1],   \
            fl_h_ice[p,t+1],    \
            fl_h_ML[p,t+1],     \
            fl_T_sfc_p[p,t+1],  \
            fl_w_dd[p,t+1],     \
            fl_u_star[p,t+1],   \
            fl_I_bot[p,t+1],    \
            fl_E[p,t+1])        \
                  = fl.Flake_Module.flake_interface( \
                        albedo_ice      = 0.9                          ,\
                        albedo_snow     = 0.9                          ,\
                        albedo_water    = albedowater                  ,\
                        c_t_in          = fl_C_T[p,t]                  ,\
                        del_time        = del_time                     ,\
                        depth_w         = depth_w                      ,\
                        dmsnowdt_in     = 0                            ,\
                        fetch           = np.sqrt(hyd_A_wb[p,t]/np.pi) ,\
                        h_ice_in        = fl_h_ice[p,t]                ,\
                        h_ml_in         = fl_h_ML[p,t]                 ,\
                        h_snow_in       = 0,
                        height_tq_in    = height_tq_in                 ,\
                        height_u_in     = height_u_in                  ,\
                        i_atm_in        = forc_I_atm[this_t]           ,\
                        p_a_in          = forc_P_a[this_t]             ,\
                        par_coriolis    = par_Coriolis                 ,\
                        q_a_in          = forc_q_a[this_t]             ,\
                        q_atm_lw_in     = forc_Q_atm_lw[this_t]        ,\
                        q_bot_in        = cr_Q_bot[p+1,t]              ,\
                        t_a_in          = T_flakeforcing               ,\
                        t_bot_in        = fl_T_bot[p,t]                ,\
                        t_ice_in        = fl_T_ice[p,t]                ,\
                        t_mnw_in        = fl_T_mnw[p,t]                ,\
                        t_sfc_p         = fl_T_sfc_p[p,t]              ,\
                        t_snow_in       = fl_T_snow[p,t]               ,\
                        t_wml_in        = fl_T_wML[p,t],    
                        u_a_in          = u_flakeforcing,
                        pondtype        = pondtype[p])

            if fl_h_ice[p,t+1] >= depth_w:
                fl_h_ice[p,t+1] = depth_w
                cr_STATVAR['T'][:int(cr_PARA['WaterDepth'][0,p+1]), p+1] = fl_T_ice[p,t+1] - 273.15
            else:
                idx_ice = (np.abs(Z - fl_h_ice[p,t+1])).argmin()
                cr_STATVAR['T'][        : idx_ice,                           p+1] = fl_T_ice[p,t+1] - 273.15
                cr_STATVAR['T'][idx_ice : int(cr_PARA['WaterDepth'][0,p+1]), p+1] = fl_T_mnw[p,t+1] - 273.15
                cr_STATVAR['T'][int(cr_PARA['WaterDepth'][0,p+1]),           p+1] = fl_T_bot[p,t+1] - 273.15
                                         
        else:
            print('too shallow - skipping FLake for waterbody #', p)
            fl_T_snow[p,t+1]= fl_T_snow[p,t]
            fl_T_ice[p,t+1] = fl_T_ice[p,t]
            fl_T_mnw[p,t+1] = forc_T_a[t]
            fl_T_wML[p,t+1] = forc_T_a[t]
            fl_T_bot[p,t+1] = fl_T_bot[p,t]
            fl_C_T[p,t+1]   = fl_C_T[p,t]
            fl_h_snow[p,t+1]= fl_h_snow[p,t]
            fl_h_ice[p,t+1] = fl_h_ice[p,t]
            fl_h_ML[p,t+1]  = depth_w
            fl_T_sfc_p[p,t+1]=fl_T_sfc_p[p,t]
            fl_w_dd[p,t+1]  = fl_w_dd[p,t]
            fl_u_star[p,t+1]= fl_u_star[p,t]
            fl_E[p,t+1]     = 0
        
          
#            ------  METHANE ------     
        ice_off = ice_on[p] >= 168 and fl_h_ice[p,t] == 0 #ice was on for a week
        if fl_h_ice[p,t] != 0:
            ice_on[p] += 1
        else:
            ice_on[p] = 0
        if meth == True:
            if fl_h_ice[p,t] > 0.2:
                this_LAI = False
            else:
                this_LAI = forc_lai[t]  
            meth_cCH4_ml_b[p,t+1]       ,\
            meth_cCH4_ml_v[p,t+1]       ,\
            meth_cO2_ml[p,t+1]          ,\
            meth_F_diff_surf_b[p,t+1]   ,\
            meth_F_diff_surf_v[p,t+1]   ,\
            meth_F_bubble_b[p,t+1]      ,\
            meth_F_bubble_v[p,t+1]      ,\
            meth_F_plant[p,t+1]         ,\
            meth_stored_CH4[p,t+1]    \
                = methane_main( \
                        H_o       = hyd_h_bare[p,t]               ,\
                        H_v       = hyd_h_veg[p,t]                 ,\
                        A_o       = hyd_A_bare[p,t] ,
                        A_v       = hyd_A_veg[p,t] ,
                        h_s       = min(0.2,cr_TD[p+1,t])                     , #Knoblauch: 20 cm much less production
                        w_dd      = fl_w_dd[p,t]         ,\
                        u_star    = fl_u_star[p,t]       ,\
                        u         = forc_U_a[t]           ,\
                        p_a       = forc_P_a[t]           ,\
                        T_ml      = fl_T_mnw[p,t]        ,\
                        T_b       = cr_T_soil[sediment_idx[p+1],p+1,t]        ,\
                        t         = del_time , \
                        ice_off   = ice_off,\
                        h_ice     = fl_h_ice[p,t]       ,\
                        del_ice   = fl_h_ice[p,t]-fl_h_ice[p,t-1],\
                        cCH4_ml_o = meth_cCH4_ml_b[p,t]    ,\
                        cCH4_ml_v = meth_cCH4_ml_v[p,t]    ,\
                        cO2_ml    = meth_cO2_ml[p,t]     ,\
                        stored_CH4= meth_stored_CH4[p,t] ,\
                        por       = por[p+1],
                        LAI       = this_LAI,
                        npp       = forc_npp[t],
                        pondtype  = pondtype[p],
                        winter=True)
                
        # ------ Hydrology ------ 
        
        # each spring reinitialize overgrown area - 0.5 water depth
        # because water depth will incresase thru season
        if t == reset_veg:
            print('resetting', forc_time[t])
            tan_a  = np.tan(topo_alpha[p])
            hyd_A_bare[p,t],\
            hyd_A_veg[p,t], \
            hyd_h_bare[p,t],\
            hyd_h_veg[p,t] = set_overgrownfraction(h_w_veg,
                                                    topo_alpha[p], 
                                                    hyd_A_wb[p,t], 
                                                    topo_super_wb[p] - hyd_WT_wb[p,t])
            print(hyd_A_bare[p,t],\
            hyd_A_veg[p,t], \
            hyd_h_bare[p,t],\
            hyd_h_veg[p,t])

          
        if fl_h_ice[p,t] > 0.1: #Snow which melted is added to the pond
            E = 0
            P = max(0, ((cr_W_soil[:air_idx+1,0,t-1]+cr_I_soil[:air_idx+1,0,t-1])
                            *cr_GRID['dxn'][:air_idx+1]).sum() 
                      - ((cr_W_soil[:air_idx+1,0,t]+cr_I_soil[:air_idx+1,0,t])
                             *cr_GRID['dxn'][:air_idx+1]).sum())
        else:
            E = fl_E[p,t]*del_time*1.E-3
            P = forc_tp[t] 
            
        hyd_A_wb[p,t+1]       ,\
        hyd_A_veg[p,t+1]      ,\
        hyd_A_bare[p,t+1]     ,\
        hyd_h_veg[p,t+1]      ,\
        hyd_h_bare[p,t+1]     ,\
        hyd_WT_wb[p,t+1]      ,\
          = hydrology(P      = P        ,
                      A_wb   = hyd_A_wb[p,t]  ,
                      A_veg  = hyd_A_veg[p,t] ,
                      A_bare = hyd_A_bare[p,t],
                      h_veg  = hyd_h_veg[p,t] ,
                      h_bare = hyd_h_bare[p,t],
                      E      = E, 
                      TD_wb  = cr_TD[p,t] ,
                      TD_rim = cr_TD[0,t] ,
                      WT_wb  = hyd_WT_wb[p,t],
                      topography_alpha = topo_alpha[p],
                      topography_super_wb = topo_super_wb[p],
                      del_time = del_time)

#%% Write output

# -- FLAKE --
flake_output = xr.Dataset({ \
                # for flake-physics
                'T_snow'    : (['waterbody','time'],fl_T_snow, {'longname':'temperature at the air-snow interface', 'unit': 'K'} ), \
                'T_ice'     : (['waterbody','time'],fl_T_ice, {'longname':'temperature at the snow-ice or air-ice interface', 'unit': 'K'} ), \
                'T_mnw'     : (['waterbody','time'],fl_T_mnw, {'longname':'mean temperature of the water column', 'unit': 'K'} ), \
                'T_wML'     : (['waterbody','time'],fl_T_wML, {'longname':'mixed-layer temperature', 'unit': 'K'} ), \
                'T_bot'     : (['waterbody','time'],fl_T_bot, {'longname':'temperature at the water-bottom sediment interface', 'unit': 'K'} ), \
                'C_T'       : (['waterbody','time'],fl_C_T, {'longname':'shape factor with respect to the temperature profile in the thermocline', 'unit': 'dimensionless'} ), \
                'h_snow'    : (['waterbody','time'],fl_h_snow, {'longname':'snow thickness', 'unit': 'm'} ), \
                'h_ice'     : (['waterbody','time'],fl_h_ice, {'longname':'ice thickness', 'unit': 'm'} ), \
                'h_ML'      : (['waterbody','time'],fl_h_ML, {'longname':'thickness of the mixed-layer', 'unit': 'm'} ), \
                'T_sfc_p'   : (['waterbody','time'],fl_T_sfc_p, {'longname':'surface temperature at the previous time step that is equal to the temperature of the surface of water, of ice or of snow' , 'unit': 'K'} ),\
                'w_dd'      : (['waterbody','time'],fl_w_dd, {'longname':'convective velocity scale, using a generalized heat flux scale', 'unit': 'm s-1'} ), \
                'u_star'    : (['waterbody','time'],fl_u_star, {'longname':'friction velocity in the surface layer of lake water', 'unit': 'm s-1'} ), \
                'I_bot'         : (['waterbody','time'],fl_I_bot, {'longname':'Radiation at pond bottom', 'unit':'W m-2 '}), \
                'E'         : (['waterbody','time'],fl_E, {'longname':'evaporation', 'unit':'kg m-2 s-1 (≈ mm s-1)'}), \
                # for flake-methane
                'cCH4_ml_v'    : (['waterbody','time'],meth_cCH4_ml_v, {'longname':'methane concentration in the mixed layer', 'unit': 'mol/m**3'} ), \
                'F_diff_surf_v': (['waterbody','time'],meth_F_diff_surf_v, {'longname':'surface diffusive methane flux to the atmosphere', 'unit': 'mol m-2 s-1'} ), \
                'F_bubble_v'   : (['waterbody','time'],meth_F_bubble_v, {'longname':'Bubble methane flux to the atmosphere', 'unit': 'mol m-2 s-1'} ), \
                'F_plant'      : (['waterbody','time'],meth_F_plant, {'longname':'Plant-mediated methane flux to the atmosphere', 'unit': 'mol m-2 s-1'} ), \
                'stored_CH4'   : (['waterbody','time'],meth_stored_CH4, {'longname':'Cummulative CH4 in water column', 'unit': 'mol m-2'} ), \
#                'ice_CH4_v'    : (['waterbody','time'],meth_ice_CH4_v, {'longname':'Cummulative CH4 in ice', 'unit': 'mol m-2'} ), \
                'cCH4_ml_b'    : (['waterbody','time'],meth_cCH4_ml_b, {'longname':'methane concentration in the mixed layer', 'unit': 'mol m-3'} ), \
                'F_diff_surf_b': (['waterbody','time'],meth_F_diff_surf_b, {'longname':'surface diffusive methane flux to the atmosphere', 'unit': 'mol m-2 s-1'} ), \
                'F_bubble_b'   : (['waterbody','time'],meth_F_bubble_b, {'longname':'Bubble methane flux to the atmosphere', 'unit': 'mol m-2 s-1'} ), \
                'cO2_ml'       : (['waterbody','time'],meth_cO2_ml, {'longname':'dissolved oxygen concentration', 'unit': 'mol m-3'} ), \
#                'ice_CH4_b'    : (['waterbody','time'],meth_ice_CH4_b, {'longname':'Cummulative CH4 in ice', 'unit': 'mol m-2'} ), \
                 }, \
                coords = {'time':forcing['time'].values, 'waterbody': topography['waterbody'].values})

flake_output.attrs = {'Forcing':Forcing_Description, 'Run': Run_Description}
flake_output.to_netcdf(Outfilename+'_FLake.nc')
 
hyd_output.attrs = {'Forcing':Forcing_Description, 'Run': Run_Description}
hyd_output.to_netcdf(Outfilename+'_hydrology.nc')

cr_output.attrs = {'Forcing':Forcing_Description, 'Run': Run_Description}
cr_output.to_netcdf(Outfilename+'_cryo.nc')

with open(Outfilename+'_cr_STRAT.pickle', 'wb') as handle:
    pickle.dump(cr_STRAT, handle, protocol=pickle.HIGHEST_PROTOCOL)

topography.attrs = {'Forcing':Forcing_Description, 'Run': Run_Description}
topography.to_netcdf(Outfilename+'_topography.nc')

forcing.to_netcdf(Outfilename+'_forcing.nc')

